use iced::{
    futures::StreamExt,
    widget::{button, checkbox, column, container, text, text_input},
    Element, Task,
};
use jid::JID;
use keyring::Entry;
use luz::{
    presence::{Offline, Presence},
    LuzHandle,
};
use serde::{Deserialize, Serialize};
use tokio_stream::wrappers::ReceiverStream;
use tracing::info;

use crate::Client;

#[derive(Default)]
pub struct LoginModal {
    jid: String,
    password: String,
    remember_me: bool,
    error: Option<Error>,
}

#[derive(Debug, Clone)]
pub enum Message {
    JID(String),
    Password(String),
    RememberMe,
    Submit,
    Error(Error),
}

#[derive(Debug, Clone)]
pub enum Error {
    InvalidJID(String),
    DatabaseConnection,
}

pub enum Action {
    None,
    ClientCreated(Task<crate::Message>),
}

#[derive(Serialize, Deserialize)]
pub struct Creds {
    pub jid: String,
    pub password: String,
}

impl LoginModal {
    pub fn update(&mut self, message: Message) -> Action {
        match message {
            Message::JID(j) => {
                self.jid = j;
                Action::None
            }
            Message::Password(p) => {
                self.password = p;
                Action::None
            }
            Message::RememberMe => {
                self.remember_me = !self.remember_me;
                Action::None
            }
            Message::Submit => {
                info!("submitting login");
                let jid_str = self.jid.clone();
                let password = self.password.clone();
                let remember_me = self.remember_me.clone();
                Action::ClientCreated(
                    Task::future(async move {
                        let jid: Result<JID, _> = jid_str.parse();
                        match jid {
                            Ok(j) => {
                                let result =
                                    LuzHandle::new(j.clone(), password.to_string(), "macaw.db")
                                        .await;
                                match result {
                                    Ok((luz_handle, receiver)) => {
                                        let mut tasks = Vec::new();
                                        tasks.push(Task::done(crate::Message::ClientCreated(
                                            Client {
                                                client: luz_handle,
                                                jid: j,
                                                connection_status: Presence::Offline(
                                                    Offline::default(),
                                                ),
                                            },
                                        )));
                                        let stream = ReceiverStream::new(receiver);
                                        let stream =
                                            stream.map(|message| crate::Message::Luz(message));
                                        tasks.push(Task::stream(stream));

                                        if remember_me {
                                            let entry = Entry::new("macaw", "macaw");
                                            match entry {
                                                Ok(e) => {
                                                    let creds = Creds {
                                                        jid: jid_str,
                                                        password,
                                                    };
                                                    let creds = toml::to_string(&creds);
                                                    match creds {
                                                        Ok(c) => {
                                                            let result = e.set_password(&c);
                                                            if let Err(e) = result {
                                                                tasks.push(Task::done(crate::Message::Error(
                                                                    crate::Error::CredentialsSave(e.into()),
                                                                )));
                                                            }
                                                        }
                                                        Err(e) => tasks.push(Task::done(
                                                            crate::Message::Error(
                                                                crate::Error::CredentialsSave(
                                                                    e.into(),
                                                                ),
                                                            ),
                                                        )),
                                                    }
                                                }
                                                Err(e) => {
                                                    tasks.push(Task::done(crate::Message::Error(
                                                        crate::Error::CredentialsSave(e.into()),
                                                    )))
                                                }
                                            }
                                        }
                                        tasks
                                    }
                                    Err(_e) => {
                                        tracing::error!("error (database probably)");
                                        return vec![Task::done(crate::Message::LoginModal(
                                            Message::Error(Error::DatabaseConnection),
                                        ))];
                                    }
                                }
                            }
                            Err(_) => {
                                tracing::error!("parsing jid");
                                return vec![Task::done(crate::Message::LoginModal(
                                    Message::Error(Error::InvalidJID(jid_str.to_string())),
                                ))];
                            }
                        }
                    })
                    .then(|tasks| Task::batch(tasks)),
                )
            }
            Message::Error(error) => {
                self.error = Some(error);
                Action::None
            }
        }
    }

    pub fn view(&self) -> Element<Message> {
        container(
            column![
                text("Log In").size(24),
                column![
                    column![
                        text("JID").size(12),
                        text_input("berry@macaw.chat", &self.jid)
                            .on_input(|j| Message::JID(j))
                            .on_submit(Message::Submit)
                            .padding(5),
                    ]
                    .spacing(5),
                    column![
                        text("Password").size(12),
                        text_input("", &self.password)
                            .on_input(|p| Message::Password(p))
                            .on_submit(Message::Submit)
                            .secure(true)
                            .padding(5),
                    ]
                    .spacing(5),
                    checkbox("remember me", self.remember_me).on_toggle(|_| Message::RememberMe),
                    button(text("Submit")).on_press(Message::Submit),
                ]
                .spacing(10)
            ]
            .spacing(20),
        )
        .width(300)
        .padding(10)
        .style(container::rounded_box)
        .into()
    }
}
